@@ -4,12 +4,13 @@
 
     The composit types for Sphinx.
 
-    :copyright: Copyright 2007-2020 by the Sphinx team, see AUTHORS.
+    :copyright: Copyright 2007-2021 by the Sphinx team, see AUTHORS.
     :license: BSD, see LICENSE for details.
 """
 
 import sys
 import typing
+from struct import Struct
 from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, TypeVar, Union
 
 from docutils import nodes
@@ -29,6 +30,11 @@ else:
             ref = _ForwardRef(self.arg)
             return ref._eval_type(globalns, localns)
 
+if sys.version_info > (3, 10):
+    from types import Union as types_Union
+else:
+    types_Union = None
+
 if False:
     # For type annotation
     from typing import Type  # NOQA # for python3.5.1
@@ -71,7 +77,9 @@ def get_type_hints(obj: Any, globalns: Dict = None, localns: Dict = None) -> Dic
         # Failed to evaluate ForwardRef (maybe TYPE_CHECKING)
         return safe_getattr(obj, '__annotations__', {})
     except TypeError:
-        return {}
+        # Invalid object is given. But try to get __annotations__ as a fallback for
+        # the code using type union operator (PEP 604) in python 3.9 or below.
+        return safe_getattr(obj, '__annotations__', {})
     except KeyError:
         # a broken class found (refs: https://github.com/sphinx-doc/sphinx/issues/8084)
         return {}
@@ -94,8 +102,17 @@ def restify(cls: Optional["Type"]) -> str:
         return ':obj:`None`'
     elif cls is Ellipsis:
         return '...'
+    elif cls is Struct:
+        # Before Python 3.9, struct.Struct class has incorrect __module__.
+        return ':class:`struct.Struct`'
     elif inspect.isNewType(cls):
         return ':class:`%s`' % cls.__name__
+    elif types_Union and isinstance(cls, types_Union):
+        if len(cls.__args__) > 1 and None in cls.__args__:
+            args = ' | '.join(restify(a) for a in cls.__args__ if a)
+            return 'Optional[%s]' % args
+        else:
+            return ' | '.join(restify(a) for a in cls.__args__)
     elif cls.__module__ in ('__builtin__', 'builtins'):
         return ':class:`%s`' % cls.__name__
     else:
@@ -305,6 +322,9 @@ def stringify(annotation: Any) -> str:
         return annotation.__qualname__
     elif annotation is Ellipsis:
         return '...'
+    elif annotation is Struct:
+        # Before Python 3.9, struct.Struct class has incorrect __module__.
+        return 'struct.Struct'
 
     if sys.version_info >= (3, 7):  # py37+
         return _stringify_py37(annotation)
@@ -329,6 +349,8 @@ def _stringify_py37(annotation: Any) -> str:
     elif hasattr(annotation, '__origin__'):
         # instantiated generic provided by a user
         qualname = stringify(annotation.__origin__)
+    elif types_Union and isinstance(annotation, types_Union):  # types.Union (for py3.10+)
+        qualname = 'types.Union'
     else:
         # we weren't able to extract the base type, appending arguments would
         # only make them appear twice
@@ -348,6 +370,12 @@ def _stringify_py37(annotation: Any) -> str:
             else:
                 args = ', '.join(stringify(a) for a in annotation.__args__)
                 return 'Union[%s]' % args
+        elif qualname == 'types.Union':
+            if len(annotation.__args__) > 1 and None in annotation.__args__:
+                args = ' | '.join(stringify(a) for a in annotation.__args__ if a)
+                return 'Optional[%s]' % args
+            else:
+                return ' | '.join(stringify(a) for a in annotation.__args__)
         elif qualname == 'Callable':
             args = ', '.join(stringify(a) for a in annotation.__args__[:-1])
             returns = stringify(annotation.__args__[-1])
